"""Statistic on used channel names.

Reason:
Add a separate table for all channel names used in any of our
conversation events. Using this table, we can properly show all channel names
without needing to iterate over all conversations.

Revision ID: 017cb274256f
Revises: b85106c82c99

"""
from collections import Counter

import json

from alembic import op
import sqlalchemy as sa
from sqlalchemy import Integer, String, column, table

# revision identifiers, used by Alembic.
revision = "017cb274256f"
down_revision = "b85106c82c99"
branch_labels = None
depends_on = None

STATISTICS_TABLE_NAME = "conversation_channel_statistic"


def _migrate_events(session):
    query = session.execute(
        "SELECT data FROM conversation_event WHERE type_name = 'user'"
    )

    used_channels = Counter()

    for row in query:
        event_data = json.loads(row[0])
        channel = event_data.get("input_channel")

        if channel is not None:
            used_channels[channel] += 1

    channel_stats_table = table(
        STATISTICS_TABLE_NAME,
        column("project_id", String),
        column("channel", String),
        column("count", Integer),
    )

    channel_counts = []

    for channel, count in used_channels.items():
        # conversation events are not attached to projects so we need to use default
        channel_counts.append(
            {"project_id": "default", "channel": channel, "count": count}
        )

    op.bulk_insert(channel_stats_table, channel_counts)


def upgrade():
    """Create a channel name statistics table and populate with data from existing events."""
    bind = op.get_bind()
    session = sa.orm.Session(bind=bind)

    op.create_table(
        STATISTICS_TABLE_NAME,
        sa.Column("project_id", sa.String(255), nullable=False),
        sa.Column("channel", sa.String(255), nullable=False),
        sa.Column("count", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(["project_id"], ["conversation_statistic.project_id"]),
        sa.PrimaryKeyConstraint("project_id", "channel"),
    )

    _migrate_events(session)


def downgrade():
    """Drop channel name statistics table."""
    op.drop_table(STATISTICS_TABLE_NAME)
