from typing import Text, Any, Dict

import sqlalchemy as sa
from sqlalchemy.orm import relationship, object_session

from rasax.community.database import utils
from rasax.community.database.base import Base
from rasax.community.database.conversation import MessageLog


class Workspace(Base):
    """Represents an NLU Workspace for training data."""

    __tablename__ = "workspace"

    id = sa.Column(sa.Integer, utils.create_sequence(__tablename__), primary_key=True)
    name = sa.Column(sa.String, nullable=False, unique=True)
    message_logs = relationship("MessageLog")
    created_at = sa.Column(sa.Float)  # annotation time as unix timestamp
    last_modified_at = sa.Column(sa.Float)  # annotation time as unix timestamp
    last_modifier_id = sa.Column(sa.String, sa.ForeignKey("rasa_x_user.username"))
    default_for_user = relationship(
        "User",
        back_populates="default_workspace",
        uselist=False,
        primaryjoin="User.default_workspace_id == Workspace.id",
    )

    def as_dict(self) -> Dict[Text, Any]:
        """Return a JSON-like representation of this `Workspace` object.

        Returns:
            Dictionary containing the workspace's attributes.
        """
        session = object_session(self)
        message_logs_count = (
            session.query(MessageLog).filter(MessageLog.workspace_id == self.id).count()
        )

        if self.default_for_user:
            default_for_username = self.default_for_user.username
        else:
            default_for_username = None

        return {
            "id": self.id,
            "name": self.name,
            "message_logs_count": message_logs_count,
            "created_at": self.created_at,
            "last_modified_by": self.last_modifier_id,
            "last_modified_at": self.last_modified_at,
            "default_for_user": default_for_username,
        }
